package farm.ants.agent;
import java.util.*;
import utilities.*;
import farm.metagent.Message;
import farm.metagent.Agent;

/**
 * An FSM controller. Handles the registration, execution, messages dispatching
 * and shared lock maintainence for PulseActionMachines that are register 
 * with the controller
 * <p>
 * Created: Thu May 24 11:43:39 2001
 * <p>
 * Modified 08/07/2002, added revoke of lock privleges and
 * changed the basic locking archtecture
 *
 * @author Roger Mailler
 */
public class PulseActionController{
    
    Hashtable Register = new Hashtable();
    Hashtable locks = new Hashtable();  //a list of the currently lock things and who has it
    Hashtable lockQueue = new Hashtable(); //the waiting lock queue keyed on lock name, elements are treemaps

    protected Log log = null;
    Agent parent = null;

    /**
     * Constructor
     * @param a The parent agent
     */
    public PulseActionController(Agent a) {
	parent = a;
	log = Log.getDefault();
    }

    /**
     * Register a new pulse action machine (PAM)
     * for the controller to handle and assigns it a random ID
     * @param PAM the new PAM to register
     * @return the ID number of the PAM
     */
    public int register(PulseActionMachine PAM){
	int newID = createID();
	Register.put(new Integer(newID), PAM);
	PAM.setID(newID);
	PAM.setController(this);
 	log.log("New machine registered " + PAM.getAction() +" ID: " + PAM.getID(), log.LOG_INFO); 
	return newID;
    }

     /**
     * Register a new pulse action machine (PAM)
     * for the controller to handle
     * @param PAM the new PAM to register
     * @param newID The ID to assign to the PAM
     * @return the ID number of the PAM
     */
    public int register(PulseActionMachine PAM, int newID){
	Register.put(new Integer(newID), PAM);
	PAM.setID(newID);
	PAM.setController(this);
 	log.log("New machine registered " + PAM.getAction() +" ID: " + PAM.getID(), log.LOG_INFO); 
	return newID;
    }

    /**
     * Unregisters a PAM from the controller
     * @param PAM the PAM to unregister
     */
    public void unregister(PulseActionMachine PAM){
	if(Register.remove(new Integer(PAM.getID()))!=null){
	    //log.log("Removed machine: " + PAM.toString(), log.LOG_DEBUG);
	}

    }

    /**
     * Handles an incoming message and dispatchs it to the correct PAM
     * all messages intented for a PAM should start with "P<ID>:"
     * @param msg The message to dispatch
     */
    public void messageReceived(Message msg){
	log.log("Got a message " + msg, log.LOG_DEBUG);

	StringTokenizer tok = new StringTokenizer(msg.getData(), ":");
	String ID = tok.nextToken();
	if(ID != null && ID.startsWith("P")){
	    Integer idNum = new Integer(ID.substring(1));
	    log.log("Checking message in PAC: " + idNum, log.LOG_DEBUG2);
	    if(Register.containsKey(idNum)){
		PulseActionMachine PAM = (PulseActionMachine) Register.get(idNum);
		log.log("Adding a message to PAM: " + idNum, log.LOG_DEBUG2);
		PAM.addMessage(msg);
	    }
	    else{
		//unkown message reciever,
		//send a message back to the sender saying unknown
		log.log("Unkown PAM " + idNum + " in in-reply-to.", log.LOG_WARNING);
	    }
	}
    }
    
    /**
     * Sends a message to another agent (Not currently used)
     * @param msg The message to send
     */
    public void sendMessage(Message msg){
	parent.sendMessage(msg);
    }
    
    /**
     * Pulses the controller who in turn pulses the registered PAMs
     * @param time The current time (system of simulator)
     */
    public void pulse(int time){
	//log.log("The pulse action controller was pulsed " + time, Log.LOG_INFO);
	Enumeration e = new SafeEnumeration(Register.elements());
	while(e.hasMoreElements()){
	    ((PulseActionMachine) e.nextElement()).pulse(time);
	}
    }

    /**
     * Requests a lock from the controller.  
     * If the lock is currently available, the PAC will call the
     * notifyLock method of the requesting PAM.  If the lock is currently
     * is use by a lower priority requester then the old lock will be revoked
     * and given to the new requester.  If the new requester is lower priority
     * then the request is placed in a lock dependent priority queue
     * @param l The PACLock object requesting the lock
     */
    public void lock(PACLock l){
	int requester = l.getRequester();
	if(locks.containsKey(l.getName())){
	    //someone currently has this lock
	    //check to see if they should have it revoked
	    PACLock currLock = (PACLock) locks.get(l.getName());
	    if(currLock.compareTo(l) > 0){
		int oldReq = currLock.getRequester();

		//the new lock has priority over the old one
		//give hime the lock
		setLock(l);

		//remove the lock from the other guy and add him to the queue
		PulseActionMachine PAM = (PulseActionMachine) Register.get(new Integer(oldReq));
		PAM.revokeLock(currLock);
		
		queueLockReq(currLock);
	    }
	    else{
		//add this request
		//to the queue
		queueLockReq(l);
	    }
	}
	else{
	    //dude, you can have it.
	    setLock(l);
	}
    }
    private void setLock(PACLock l){
	locks.put(l.getName(), l);
	PulseActionMachine PAM = (PulseActionMachine) Register.get(new Integer(l.getRequester()));
	PAM.notifyLock(l);
    }

    private void queueLockReq(PACLock l){
	TreeSet queue = null;
	if(lockQueue.containsKey(l.getName())){
	    queue = (TreeSet) lockQueue.get(l.getName());
	}
	else{
	    queue = new TreeSet();
	}
	queue.add(l);
	lockQueue.put(l.getName(), queue);
    }

    /**
     * Tells the controller that a machine no longer needs a lock.
     * If there are other machines waiting on the lock then
     * they are given the lock in priority order according to there PACLock requests.
     * If the lock request is in the queue, then the queued request is removed
     * @param l the PACLock currently possessing the lock
     */
    public void unlock(PACLock l){
	int requester = l.getRequester();
	if(locks.containsKey(l.getName())){
	    if(l.equals((PACLock) locks.get(l.getName()))){
		locks.remove(l.getName());
		if(lockQueue.containsKey(l.getName())){
		    //check to see if anyone is in the queue
		    //waiting
		    TreeSet queue = (TreeSet) lockQueue.get(l.getName());
		    if(queue.size()> 0){
			PACLock PACL = (PACLock) queue.first();
			setLock(PACL);
			queue.remove(PACL);
		    }
		}
	    }
	    else{
		//it didn't have the lock, but may want to
		//be removed from the queue
		if(lockQueue.containsKey(l.getName())){
		    TreeSet queue = (TreeSet) lockQueue.get(l.getName());
		    queue.remove(l);
		}		
	    }
	}
    }

    /**
     * Finds out whether or not a lock is currently being used
     * @param lockName the lock name to check
     * @return true if the lock is in use
     */
    public boolean isLocked(String lockName){
	return locks.containsKey(lockName);
    }

    /**
     * Finds out if a particular machine ha a particular lock
     * @param lockName the name of the lock
     * @param requester the ID number of the machine
     * @return true if that machine has the lock
     */
    public boolean hasLock(String lockName, int requester){
	if(locks.containsKey(lockName)){
	    if(((PACLock) locks.get(lockName)).getRequester() == requester){
		return true;
	    }
	}
	return false;
    }

    /**
     * Returns the number of locks that the requester has
     * @param requester the ID number of the requester
     * @return the number of locks owned by the requester
     */
    public int numLocks(int requester){
	int total = 0;
	for(Iterator i = locks.values().iterator();i.hasNext();){
	    if(((PACLock) i.next()).getRequester() == requester){
		total++;
	    }
	}
	return total;
    }

    /**
     * Gets a list of locks currented possesed by the requester
     * @param requester the ID number of the requester
     * @return a Vector of the locks owned by the requester
     */
    public Vector getLocks(int requester){
	Vector ret = new Vector();
	for(Iterator i = locks.keySet().iterator();i.hasNext();){
	    String lockName = (String) i.next();
	    if(((PACLock) locks.get(lockName)).getRequester() == requester){
		ret.add(lockName);
	    }
	}
	return ret;	
    }

    /**
     * Gets a list of the PACLocks that currently have locks
     * @return a vector of PACLock objects
     */
    public Vector getCurrLocks(){
	return new Vector(locks.keySet());
    }

    /**
     * Prints a list of the PACLocks that currently have locks
     */
    public void printLocks(){
	log.log("Current Locks", log.LOG_INFO);
	for(Iterator i = locks.values().iterator();i.hasNext();){
	    log.log(((PACLock) i.next()).toString());
	}
    }

    private int createID(){
   	Random tmpR = new Random();
	//set up a new state machine to process the alteration
	int newID = tmpR.nextInt(100000);
	while(Register.containsKey(new Integer(newID)))
	     newID = tmpR.nextInt(100000);
	return newID;
    }



} // PulseActionController
