package farm.ants.agent;
import java.util.*;

/**
 * This is a state maintaince class for a resource focused SPAM agent.
 * Primarliy, it maintains a cache of resource states, the local assignment
 * of resources, a list of conflictes resources, the local objective function
 * and level, and the context of the last resource change 
 * (may have had a negotiation to resolve it)
 * <p>
 * Created: Mon Sep 11 12:02:12 2000
 * <p>
 * Migrated from ANTS to farm 07/16/2002 by Roger
 *
 * @author Roger Mailler
 */
public class Track  {
    Hashtable states = new Hashtable();   //used to hold infomration about 
                                          //my allocation for this sensor

    Hashtable schedules = new Hashtable();  //caches the schedules of the
                                            //agents that are tracking

    Vector conSensors = new Vector();  //a list of the sensors that I current 
                                       //am using that are under conflcit 
                                       //with another track manager

    Vector context = new Vector();    //context maintains information
    //about how the current track was derived.  context holds the
    //names of the agents that were under consideration before the
    //selection was made that is passed to SPAM


    ObjFunction objective = new SPAMObjFunction();  //the function for objective functions


    /**
     * Constructor
     */
    public Track(){
    }
    
    /******************************************/
    /*      Assignment Accessors              */
    /******************************************/

    /**
     * Gets a list of agents that are currenly being used to track
     * @return The list (Vector of strings)
     */
    public Vector getAgents(){
	return new Vector(states.keySet());
    }

    /**
     * Adds an agent to the list of agents being used
     * @param name The name of the agent
     * @param state The assigned schedule for the agent
     */
    public void addAgent(String name, SlotSchedule state){
	states.put(name, state);
    }

    /**
     * Removes an agent from the currently using list
     * @param name The name of the agent
     */
    public void removeAgent(String name){
	states.remove(name);
	schedules.remove(name);
    }

    /**
     * Removes all agent from the currently using list.
     * Also removes any cached schedule they may have.
     */
    public void clearAgents(){
	states.clear();
	schedules.clear();
    }

    /**
     * Returns the number of agents currently being used to track
     * @return The number of agents
     */
    public int numAgents(){return states.size();}

    /**
     * Determines if a particular agent is being used
     * @param name The name of the agent
     * @return True if it is being used
     */
    public boolean hasAgent(String name){return states.containsKey(name);}

    /**
     * Sets the schedule of an agent in the used list
     * @param name The name of the agent
     * @param state The assigned schedule for the agent
     */
    public void setAgentState(String name, SlotSchedule state){
	states.put(name,state);
    }

    /**
     * Gets the list of all agents being used and their state
     * @return Hashtable keyed on name, of SlotSchedules
     */
    public Hashtable getAssScheds(){
	Hashtable HT = new Hashtable();
	for(Iterator i = states.keySet().iterator();i.hasNext();){
	    String name = (String) i.next();
	    SlotSchedule SS = (SlotSchedule) states.get(name);
	    HT.put(name, SS);
	}
	return HT;
    }


    /******************************************/
    /*      Objective function  Accessors     */
    /******************************************/ 

    /**
     * Gets the objective function for the agent
     * @return The objective function
     */
    public ObjFunction getObjFunc(){
	return objective;
    }

    /******************************************/
    /*      Context Accessors                 */
    /******************************************/ 

    /**
     * Sets the context of the last resource change
     * @param c A vector of SensorData describing the context
     */
    public void setContext(Vector c){
	context = (Vector) c.clone();
    }

    /**
     * Gets the context of the last resource change
     * @return A vector of SensorData describing the context
     */
    public Vector getContext(){
	return context;
    }

    /**
     * Checks to see if a particular resource is in the context
     * @param name the resource name
     * @return True if it is
     */
    public boolean inContext(String name){
	for(Iterator i = context.iterator();i.hasNext();){
	    if(((SensorData) i.next()).getName().equals(name))
		return true;
	}
	return false;
    }

    /**
     * Checks to see if the current and past resource contexts are the same
     * @param other The new resource context
     * @return True if it is
     */
    public boolean contextEquals(Vector other){
	//three checks first size, then balance both ways

	if(context.size() != other.size()) return false;
	for(Iterator i = context.iterator();i.hasNext();)
	    if(!other.contains(i.next()))
		return false;
	
	//for(Iterator i = other.iterator();i.hasNext();)
	//   if(!context.contains(i.next()))
	//	return false;
	
	return true;
    }

    /******************************************/
    /*      Schedule Accessors                */
    /******************************************/ 

    /**
     * Adds a schedule to the schedule cache.
     * @param name of the sensor the schedule belongs to.
     * @param the schedule for the sensor.
     */
    public void addSchedule(String name, SlotSchedule SS){
	schedules.put(name, SS.clone());
    } 

    /**
     * Gets a chached sensor schedule.
     * @param the name of the sensor
     * @return schedule for the sensors.
     */
    public SlotSchedule getSchedule(String name){
	return (SlotSchedule) schedules.get(name);
    }

    /**
     * Determines if a schedule exists in the cache for a particular sensor.
     * @param the name of the sensor
     * @return true if the schedule exists.
     */    
    public boolean hasSchedule(String name){
	return schedules.containsKey(name);
    }

    /**
     * Get the list of cached schedules.
     * @return the cached schedules.
     */
    public Hashtable getSchedules(){
	return schedules;
    }

    /**
     * Empties the schedule cache.
     */
    public void clearSchedules(){
	schedules.clear();
    }

    /**
     * Empties the schedule cache for a particular agent.
     * @param the name of the agent
     */
    public void clearSchedule(String name){
	schedules.remove(name);
    }

    /******************************************/
    /*      conSensors Accessors              */
    /******************************************/ 

    /**
     * Get the value of conSensors.
     * @return Value of conSensors.
     */
    public Vector getConSensors(){return conSensors;}
    
    /**
     * Set the value of conSensors.
     * @param cons  Value to assign to conSensors.
     */
    public void setConSensors(Vector cons){conSensors = cons;}
    
    /**
     * Is the sensor on the conflict list.
     * @return true is sensor is in list.
     */    
    public boolean isConSensor(String name){return conSensors.contains(name);}
    
} // Track
