package farm.bandit.agent;

import farm.simulator.*;
import farm.metagent.*;

import java.util.*;
import utilities.Log;

/**
 * This implements a learning algorithm based on the 
 * REINFORCE paradigm.  Specifically, it uses an adaptive
 * baseline value which gives more weight to an arms average
 * performance over individual returns.  These baselines are
 * used to maintain preference parameters, which are
 * interpreted using a Gibb's distribution to obtain probabilities.
 */
public class ReinforceAgent extends BanditAgent {
    public static String NAME = "rdm";
    protected double alpha = 0.1;
    protected double p[];
    protected double pi[];
    protected double r[];
    protected double rbar;
    protected double x = 0.1;
    protected double y = 1;

    /**
     * Constructor
     * @param metagent A handle to the metagent this agent belongs to
     * @param name The agent's name
     */
    public ReinforceAgent(FarmMetAgent metagent, String name) {
	super(metagent, name);
    }

    /**
     * Constructor
     * @param metagent A handle to the metagent this agent belongs to
     * @param alpha The alpha value to use
     * @param x The x value to use
     * @param y The y value to use
     * @param name The agent's name
     */
    public ReinforceAgent(FarmMetAgent metagent, String name, double alpha, double x, double y) {
        this(metagent, name);

        this.alpha = alpha;
        this.x = x;
        this.y = y;
    }

    /**
     * Initialize the preference, probability and average
     * reward state vectors.
     */
    public void begin() {
        super.begin();

        p = new double[getNumArms()];
        pi = new double[getNumArms()];
        r = new double[getNumArms()];
        rbar = 0;
        for (int i = 0; i < p.length; i++) {
            p[i] = 0;
            r[i] = 0;
        }
    }

    /**
     * Called once for each play
     */
    public void pulse() {
        super.pulse();
        int arm = 0;

        // Compute the probabilities for the arms
        for (int i = 0; i < pi.length; i++) {
            pi[i] = pi(i);
        }

        // First pick an arm
        float c = random.nextFloat();
        for (int i = 0; i < pi.length; i++) {
            arm = i;
            if (c < pi[arm]) {
                break;
            }
            c -= pi[i];
        }

        // Then pull it
        double reward = pullArm(arm);

        // Calculate the baseline, using rbar or r[arm] depending
        // on how long we have been tracking the values
	float k = getNumPulls(arm);
        double b = (rbar
                + reward * (1 - 1d / Math.pow(k, x)) 
                - r[arm] * (1 - 1d / Math.pow(k, y))
        );

        // Update the delta value
        double delta = alpha * (reward - b) * (1 - pi[arm]);

        // Update the arm's preference
        p[arm] = p[arm] + delta;

        // Update r, the arm's average reward
        r[arm] = rbar + (1f / k) * (reward - r[arm]);

        // Update rbar, the overall average reward
        rbar = rbar + (1f / (float)getNumPulls()) * (reward - rbar);

        //System.err.println(getName() + " Arm:" + arm + " Reward:" + reward + " Optimal:" + isOptimalArm(arm) + " p:" + p[arm] + " pi:" + pi[arm] + " rbar:" + rbar + " d:" + d[arm]);
    }

    /**
     * Calculates the probability for a particular arm using
     * a Gibb's distribution
     * @param a The arm in question
     * @return The probability that arm should be selected
     */
    public double pi(int a) {
        double n = Math.exp(p[a]);
        double d = 0;

        for (int i = 0; i < p.length; i++) {
            d += Math.exp(p[i]);
        }

        return (n / d);
    }
}
