package farm.generic.plugins;

import java.rmi.*;
import java.rmi.server.*;
import farm.simulator.*;

import java.util.*;
import java.io.*;

import utilities.Log;

/**
 * This utility plugin can be used to continuously log specified
 * properties to a file.  A list of properties is provided with the
 * -log parameter.  Each time the plugin is pulse, each element in
 * that list is passed to the findProperties function, the results
 * of which are used to fetch the associated values.  All returned
 * values are then logged to the output file with their toString
 * function.
 * <p>
 * Note that because findProperties is used (as opposed to getProperty)
 * the specified property names are implicit search terms, so that
 * if "Bar" were given as a property to log, the values for "Bar", "FooBar",
 * and "BarFoo" would all be recorded.
 * <p>
 * Parameters:<dl>
 * <dt><tt>-o [string]</tt>
 * <dd>The output file's name
 *
 * <dt><tt>-log [comma separated list]</tt>
 * <dd>A list of properties to log
 *
 * <dt><tt>-format [normal, column]</tt>
 * <dd>Whether output should be normal (one line per time and property) or 
 *     columner (one line per property) default normal
 *
 * </dl>
 */
public class PropertyLog extends FarmPlugIn {

    // Constants
    public static final String CLA_LOG = "log";
    public static final String CLA_OUTPUT = "o";
    public static final String CLA_FORMAT = "format";
    public static final String CLA_FORMAT_NORMAL = "normal";
    public static final String CLA_FORMAT_COLUMN = "column";

    // Local information
    protected Vector track = new Vector();
    protected BufferedWriter output = null;
    protected String outFormat = null;
    protected boolean startup = true; 
    /**
     * Constructor
     * @param address The address of the remote simulator to eonnect to
     */
    public PropertyLog(String []args)throws RemoteException {
        super(args);

        // Command line args
        setPlugInName(getCommandLineArgument(args, CLA_NAME, "PropertyLog"));
        setOutputName(getCommandLineArgument(args, CLA_OUTPUT, null));
        setOutputFormat(getCommandLineArgument(args, CLA_FORMAT, CLA_FORMAT_NORMAL));
	
        String totrack = getCommandLineArgument(args, CLA_LOG, null);
        if (totrack != null) {
            for (StringTokenizer tok = new StringTokenizer(totrack, ","); tok.hasMoreTokens(); ) {
                String t = tok.nextToken();
                logData(t);
            }
        }
    }

    /**
     * Specifies and opens the file which should be used to log
     * property information to
     * @param name The name of the file to log to
     */
    public void setOutputName(String name) {
        if (name == null) return;

        try {
            FileWriter w = new FileWriter(name);
            output = new BufferedWriter(w);
        } catch (IOException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }

    /**
     * Specifies the format for data output
     * @param type either normal of column
     */
    public void setOutputFormat(String type) {
        if (type.equals(CLA_FORMAT_COLUMN))
	    outFormat = type;
	else
	    outFormat = CLA_FORMAT_NORMAL;
    }

    /**
     * Indicates that the specified data name should be logged.
     * The value will be retrieved during the pulseEnd phase, and
     * logged with it's toString function during each pulse
     * @param name The name of the property key to log
     */
    public void logData(String name) {
       
        synchronized (track) {
            log.log("Logging " + name, Log.LOG_INFO);
            track.addElement(name);
        }
    }

    /**
     * This is called at the end of a pulse.
     */
    protected void pulseEnd() throws RemoteException {

        // Log the data
        try {
	    if (startup && outFormat.equals(CLA_FORMAT_COLUMN)){
		//print the file header
		output.write("Time\t");
                for (Iterator i = track.iterator(); i.hasNext(); ) {
		    output.write((String) i.next());
		    if(i.hasNext())
			output.write("\t");
		    else
			output.write("\n");			
		}		
		startup = false;
	    }

            if (timeChanged()) {
                int time = getTime();
		if(outFormat.equals(CLA_FORMAT_NORMAL)){
		    for (Iterator i = track.iterator(); i.hasNext(); ) {
			String name = (String)i.next();
                        Collection array = findProperties(name);
			for (Iterator ai = array.iterator(); ai.hasNext(); ) {
                            String prop = (String)ai.next();
			    Object value = getProperty(prop);
			    output.write(time + "\t" + prop + "\t" + value + "\n");
			}
		    }
		}
		else {
		    output.write(time + "\t");
		    for (Iterator i = track.iterator(); i.hasNext(); ) {
			String name = (String)i.next();
			Object value = getProperty(name);
			output.write("" + value);
			if(i.hasNext())
			    output.write("\t");
			else
			    output.write("\n");		
		    }
		}
            }
            output.flush();
        } catch (IOException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }

    /**
     */
    public void end() throws RemoteException {

        // Deal with the file
        if (output != null) {
            try {
                output.flush();
                output.close();
            } catch (IOException e) {
                System.err.println("Exception: " + e.getMessage());
                e.printStackTrace();
            }
        }

        super.end();
    }

    /**
     * The main function
     */
    public static void main(String[] args) {

        try {
            FarmPlugIn plugin = new PropertyLog(args);
            plugin.run();
        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }
}
