package farm.generic.plugins.gui;

import java.rmi.*;
import java.rmi.server.*;
import farm.simulator.*;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.text.*;

import utilities.Log;

/**
 * This is a general graph interface, which can be used to display specified
 * numeric properties stored in the simulator.  The easiest way to use this
 * plugin is to use the -track parameter, with which you may specify a
 * comma-delimited list of property names to graph at runtime.  For example
 * <pre>
 *    java FarmPlugIn -track "Foo,Bar,Bin"
 * </pre>
 * ...would query the simulator for the value of the three Parameters Foo, Bar
 * and Bin each time it was pulsed, store those values in a local data table,
 * and then graph the results in its window.
 * <p>
 * Parameters:<dl>
 * <dt><tt>-track [comma separated list]</tt>
 * <dd>A list of properties to track and graph
 * <dt><tt>-timegranularity [integer]</tt>
 * <dd>The time granularity to use when drawing the x-axis
 * </dl>
 */
public class GraphGui extends FarmPlugIn {

    // Constants
    public static final int GRAPH_LINE = 0;
    public static final int GRAPH_COLUMN = 1;

    public static final String CLA_TRACK = "track";
    public static final String CLA_TIMEGRANULARITY = "timegranularity";

    // Local information
    protected JFrame window;
    protected GraphPanel graphpanel;
    protected Vector track = new Vector();
    protected Hashtable datakeys = new Hashtable();
    protected double maxy = Double.MIN_VALUE;
    protected double miny = 0;
    protected int timegran = 100;

    // The data table is a Hashtable of TreeMaps, keyed by the dataname.
    // Each treemap is keyed with the time of the data, which references
    // the data observed at that time.
    protected TreeMap data = new TreeMap();

    /**
     * Constructor
     * @param address The address of the remote simulator to eonnect to
     */
    public GraphGui(String []args)throws RemoteException {
        super(args);

        // Command line args
        setPlugInName(getCommandLineArgument(args, CLA_NAME, "GraphGui"));

        timegran = Integer.parseInt(getCommandLineArgument(args, CLA_TIMEGRANULARITY, String.valueOf(timegran)));

        String totrack = getCommandLineArgument(args, CLA_TRACK, null);
        if (totrack != null) {
            for (StringTokenizer tok = new StringTokenizer(totrack, ","); tok.hasMoreTokens(); ) {
                String t = tok.nextToken();
                trackData(t);
            }
        }
    }

    /**
     * Indicates that the specified data name should be tracked.
     * The value will be retrieved during the pulseEnd phase, and
     * stored as a Double in the data table.  This is just a
     * convenience function, you can always call addData directly.
     * <p>
     * Note that you can use the command line argument "track" set
     * to a comma-delimited list to have the grapher automatically
     * track a specified list of properties.
     * @param name The name of the property key to track
     */
    public void trackData(String name) {
        
        synchronized (track) {
            log.log("Tracking " + name, Log.LOG_INFO);
            track.addElement(name);
        }
    }

    /**
     */
    public void begin() throws RemoteException{
        super.begin();

        // Make and show the window
        window = new JFrame("Graph");
        window.getContentPane().add(graphpanel = new GraphPanel(this, timegran));
        window.setSize(640,400);
        window.setVisible(true);
        window.pack();

        // Listen
        window.addWindowListener(new WindowAdapter() {
                public void windowClosing(WindowEvent e) {
                    setStop();
                }
        });
    }

    /**
     * This is called at the end of a pulse.
     */
    protected void pulseBegin() throws RemoteException {

        System.err.println("Pulse begin " + System.currentTimeMillis());
    }

    /**
     * This is called at the end of a pulse.
     */
    protected void pulseEnd() throws RemoteException {

        // Look for data
        System.err.println("Pulse end " + System.currentTimeMillis());
        if (timeChanged()) {
            int time = getTime();
            for (Iterator i = track.iterator(); i.hasNext(); ) {
                String name = (String)i.next();
                Collection array = findProperties(name);
                for (Iterator ai = array.iterator(); ai.hasNext(); ) {
                    String prop = (String)ai.next();
                    Object value = getProperty(prop);
                    if (value instanceof Number) {
                        addData(prop, time, (Number)value);
                    } else if (value != null) {
                        log.log("Could not track property " + prop + ", it is not a Number", Log.LOG_WARNING);
                    }
                }
            }
        }
        System.err.println("Pulse complete " + System.currentTimeMillis());
    }

    /**
     */
    public void end() throws RemoteException {

        // Deal with the window
        window.dispose();

        super.end();
    }

    /**
     * Adds a new data value to the list
     */
    public void addData(String name, int time, Number value) {
        TreeMap map;

        if (value == null)
            return;
        if (Float.isNaN(value.floatValue()))
            return;

        map = (TreeMap)data.get(name);
        if (map == null) {
            map = new TreeMap();
            synchronized (data) {
                data.put(name, map);
            }
            addDataKey(name);
        }

        synchronized (map) {
            map.put(new Integer(time), value);
        }

        if (isDataKeyEnabled(name)) {
            // Determine the new bounds
            double n = value.doubleValue();
            if (n > maxy) {
                maxy = Math.ceil(n);
            }
            if (n < miny) {
                miny = Math.floor(n);
            }

            // Update the display
            if (graphpanel != null) {
                graphpanel.updateDisplay();
            }
        }
    }

    /**
     * Recomputes the minimum and maximum values for all
     * enabled data sets
     */
    public void recomputeDataBounds() {

        maxy = Double.MIN_VALUE;
        miny = 0;
        
        for (Iterator e = getDataKeys(); e.hasNext(); ) {
            String name = (String)e.next();
            if (!isDataKeyEnabled(name)) continue;
            TreeMap map = (TreeMap)data.get(name);
            for (Iterator i = map.entrySet().iterator(); i.hasNext(); ) {
                Map.Entry entry = (Map.Entry)i.next();
                double n = ((Number)entry.getValue()).doubleValue();
                if (n > maxy) {
                    Math.ceil(n);
                }
                if (n < miny) {
                    miny = Math.floor(n);
                }
            }
        }
    }

    /**
     * Gets the data container.  Don't modify the contents.
     */
    public TreeMap getData() {

        return data;
    }

    /**
     * Returns an iterator over the TreeMap entries in the
     * data container.
     */
    public Iterator getDataEntries() {
        Vector maps = new Vector();

        synchronized (data) {
            for (Iterator i = data.entrySet().iterator(); i.hasNext(); ) {
                TreeMap map = (TreeMap)((Map.Entry)i.next()).getValue();
                maps.addElement(map);
            }
        }

        return maps.iterator();
    }

    /**
     * Returns an iterator over the TreeMap keys in the
     * data container.
     */
    public Iterator getDataKeys() {
        Vector keys = new Vector();

        synchronized (data) {
            for (Iterator i = data.keySet().iterator(); i.hasNext(); ) {
                keys.addElement(i.next());
            }
        }

        return keys.iterator();
    }

    /**
     * Key enablement/disablement accesors
     */
    public void addDataKey(String key) {
        enableDataKey(key);
    }
    public void enableDataKey(String key) {

        datakeys.put(key, Boolean.TRUE);

        if (graphpanel != null) {
            recomputeDataBounds();
            graphpanel.updateDisplay();
        }
    }
    public void disableDataKey(String key) {

        datakeys.put(key, Boolean.FALSE);

        if (graphpanel != null) {
            recomputeDataBounds();
            graphpanel.updateDisplay();
        }
    }
    public boolean isDataKeyEnabled(String key) {

        Boolean b = (Boolean)datakeys.get(key);
        if (b != null)
            return b.booleanValue();
        else
            return false;
    }

    /**
     * Gets the minimum/maximum y value for the data, useful for scaling
     */
    public double getMaximumYValue() {
        return maxy;
    }
    public double getMinimumYValue() {
        return miny;
    }

    /**
     * Determines the start time of the data
     */
    public int getStartTime() {
        int time = Integer.MAX_VALUE;
        Vector maps = new Vector();

        for (Iterator i = getDataEntries(); i.hasNext(); ) {
            TreeMap map = (TreeMap)i.next();
            synchronized (map) {
                if (map.size() > 0) {
                    time = Math.min(time, ((Integer)map.firstKey()).intValue());
                }
            }
        }
        
        return (time != Integer.MAX_VALUE) ? time : 0;
    }

    /**
     * Determines the finish time of the data
     */
    public int getFinishTime() {
        int time = 0;

        for (Iterator i = getDataEntries(); i.hasNext(); ) {
            TreeMap map = (TreeMap)i.next();
            synchronized (map) {
                if (map.size() > 0) {
                    time = Math.max(time, ((Integer)map.lastKey()).intValue());
                }
            }
        }
        
        return time;
    }

    /**
     * The main function
     */
    public static void main(String[] args) {

        try {
            FarmPlugIn plugin = new GraphGui(args);
            plugin.run();
        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        } catch (Throwable e) {
            System.err.println("Error: " + e.getMessage());
            e.printStackTrace();
        }
    }
}

/**
 * The graph window
 */
class GraphPanel extends JPanel {
    JScrollPane graphscroll;
    JPanel graphcanvas;
    JPanel timecanvas;
    JPanel unitscanvas;
    JPanel legendcanvas;

    GraphGui gui;

    public float TICKWIDTH = 0.1f;
    public float TICKINC = 50 / TICKWIDTH;
    private int PAD = 0;
    private int HINSET = PAD * 4;
    private int TIME_GRANULARITY = 1;

    protected boolean resizescroll = true;
    protected int style = GraphGui.GRAPH_LINE;

    protected static Color[] colors;
    protected static Hashtable colorset = new Hashtable();
    static {
        colors = new Color[8];
        colors[0] = new Color(255, 0, 0);  // red
        colors[1] = new Color(0, 153, 0);  // green
        colors[2] = new Color(0, 0, 255);  // blue
        colors[3] = new Color(153, 0, 204); // purple
        colors[4] = new Color(102, 102, 0); // brown
        colors[5] = new Color(204, 102, 0); // orange
        colors[6] = new Color(255, 51, 255); // magenta
        colors[7] = new Color(102, 102, 153); // slate
    }

    public GraphPanel(GraphGui g, int tg) {
        super();
        
        TIME_GRANULARITY = tg;

        // Save the reference
        gui = g;

        // Set up the graph
        graphscroll = new JScrollPane();
        graphscroll.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
        graphscroll.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_NEVER);
        graphcanvas = new GraphCanvas(gui, graphscroll);
        graphscroll.getViewport().add(graphcanvas);
        timecanvas = new TimelineCanvas(gui, graphscroll);
        graphscroll.setColumnHeaderView(timecanvas);
        unitscanvas = new UnitsCanvas(gui, graphscroll);
        legendcanvas = new LegendCanvas(gui, graphscroll);
        JPanel panel = new JPanel();
        panel.setLayout(new BorderLayout());
        panel.add(legendcanvas, BorderLayout.WEST);
        panel.add(unitscanvas);
        graphscroll.setRowHeaderView(panel);

        // Set up the prefs
        JPanel prefspanel = new JPanel();
        JComboBox stylemenu = new JComboBox();
        stylemenu.addItem("Lines");
        stylemenu.addItem("Columns");
        stylemenu.setSelectedIndex(style);
        JCheckBox updatescroll = new JCheckBox("Scroll on update");
        updatescroll.setSelected(resizescroll);
        JSlider timeslider = new JSlider(1, 50, (int)(TICKWIDTH * 100));
        prefspanel.add(new JLabel("Style:"));
        prefspanel.add(stylemenu);
        prefspanel.add(updatescroll);
        prefspanel.add(timeslider);

        // Place them
        setLayout(new BorderLayout());
        add(graphscroll);
        add(prefspanel, BorderLayout.SOUTH);

        // Listen to their stories
        graphcanvas.addComponentListener(new ComponentAdapter() {
                public void componentResized(ComponentEvent e) {
                    if (resizescroll) {
                        JScrollBar scroll = graphscroll.getHorizontalScrollBar();
                        scroll.setValue(scroll.getMaximum());
                    }
                }
            });
        updatescroll.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    resizescroll = ((JCheckBox)e.getSource()).isSelected();
                }
            });
        timeslider.addChangeListener(new ChangeListener() {
                public void stateChanged(ChangeEvent e) {
                    int value = ((JSlider)e.getSource()).getValue();
                    TICKWIDTH = (float)value / 100f;
                    TICKINC = 50 / TICKWIDTH;
                    updateDisplay();
                }
            });
        stylemenu.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    style = ((JComboBox)e.getSource()).getSelectedIndex();
                    updateDisplay();
                }
            });

        // Background colors
        setBackground(Color.white);
        graphscroll.setBackground(Color.white);
    }

    public void drawString(Graphics g, String s, int x, int y, int align, int valign) {
        FontMetrics metrics = g.getFontMetrics();
        int width = metrics.stringWidth(s);
        int height = metrics.getAscent();
        
        y -= metrics.getDescent();
        
        switch (align) {
        case 0: // Left
            break;
        case 1: // Center
            x -= width / 2;
            break;
        case 2: // Right
            x -= width;
            break;
        }
        
        switch (valign) {
        case 0: // Bottom
            break;
        case 1: // Center
            y += height / 2;
            break;
        case 2: // Top
            y += height;
            break;
        }

        g.drawString(s, x, y);
    }

    public void updateDisplay() {
        timecanvas.invalidate();
        graphcanvas.invalidate();
        unitscanvas.invalidate();
        legendcanvas.invalidate();
        revalidate();
        repaint();
    }

    public int calculateTime(int x) {
        int time = Math.round(((float)x + (TICKWIDTH / 2.0f)) / TICKWIDTH) + getStartTime();
        if (time < getStartTime() || time > getFinishTime())
            return -1;
        else
            return time;
    }

    public int getStartTime() {
        int time = gui.getStartTime();
        return time - (time % (int)TICKINC);
    }

    public int getFinishTime() {
        int time = gui.getFinishTime();
        return time + (int)TICKINC - (time % (int)TICKINC);
    }

    /**
     * Gets a specific perty color, for use by the visualization part.
     * In general, only numbers 0-2 (red, green, blue) will remain
     * constant, any past that may be changed in later revisions..
     */
    public static Color getColor(int i) {
        return colors[i];
    }

    /**
     * Gets a perty color, for use by the visualization part.
     */
    public static Color getColor(String s) {
        Integer color = (Integer)colorset.get(s);
        if (color != null) return colors[color.intValue()];

        // Pick a color
        Vector v;
        color = new Integer(0);
        for (int i = 0; i < colors.length - 1; i++) {
            Integer num1 = (Integer)colorset.get(colors[i]);
            if (num1 == null) {
                num1 = new Integer(0);
                colorset.put(colors[i], num1);
            }
            Integer num2 = (Integer)colorset.get(colors[i+1]);
            if (num2 == null) {
                num2 = new Integer(0);
                colorset.put(colors[i+1], num2);
            }
            if (num2.intValue() < num1.intValue()) {
                color = new Integer(i+1);
                break;
            }
        }
        Integer num = (Integer)colorset.get(colors[color.intValue()]);
        colorset.put(colors[color.intValue()], new Integer(num.intValue() + 1));
        colorset.put(s, color);

        return colors[color.intValue()];
    }

    /**
     * The graph panel
     */
    class TimelineCanvas extends JPanel {
        JScrollPane graphscroll;
        GraphGui gui;

        int TIME_HEIGHT = 25;
        int TIME_LINE = TIME_HEIGHT - 5;
        int TIME_TEXT = TIME_LINE - 5;

        public TimelineCanvas(GraphGui g, JScrollPane scroll) {
            gui = g;
            graphscroll = scroll;
            setAutoscrolls(true);
        }

        public boolean isOpaque() { return true; }
        // 1.3
        public boolean isFocusTraversable() { return true; }
        // 1.4
        //public boolean isFocusable() { return true; }

        public void paintComponent(Graphics g) {
            int start = getStartTime();
            int finish = getFinishTime();
            Rectangle bounds = graphscroll.getViewport().getViewRect();
            Color c;
            
            // Limit what we actually draw
            int vstart = calculateTime(bounds.x);
            if (vstart < 0) vstart = start;
            int vfinish = calculateTime(bounds.x + bounds.width);
            if (vfinish < 0) vfinish = finish;

            // Blank
            g.setColor(Color.white);
            g.fillRect(bounds.x, bounds.y, bounds.width, bounds.height);

            // Draw the timeline
            vstart -= vstart % TICKINC;
            vfinish += TICKINC + TICKINC - (vfinish % TICKINC);
            g.setColor(Color.black);
            g.drawLine(Math.round((vstart - start) * TICKWIDTH), TIME_LINE,
                       Math.round((vfinish - start) * TICKWIDTH), TIME_LINE);
            for (int i = vstart; i < vfinish; i += TICKINC) {
                g.drawLine(Math.round((i - start) * TICKWIDTH), TIME_LINE,
                           Math.round((i - start) * TICKWIDTH), TIME_LINE + 2);
                drawString(g, String.valueOf(i / TIME_GRANULARITY),
                           Math.round((i - start) * TICKWIDTH), TIME_TEXT, 1, 0);
            }
        }

        public Dimension getPreferredSize() {
            int w = Math.max(Math.round((getFinishTime() - getStartTime()) * TICKWIDTH), 350);
            return new Dimension(w, TIME_HEIGHT);
        }
        public Dimension getMinimumSize() { return getPreferredSize(); }
        public Dimension getMaximumSize() { return getPreferredSize(); }
    }

    /**
     * The graph panel
     */
    class UnitsCanvas extends JPanel {
        JScrollPane graphscroll;
        GraphGui gui;
        NumberFormat format;

        int UNITS_WIDTH = 75;
        int UNITS_LINE = UNITS_WIDTH - 5;
        int UNITS_TEXT = UNITS_LINE - 5;
        int UNITS_DIV = 25;

        public UnitsCanvas(GraphGui g, JScrollPane scroll) {
            gui = g;
            graphscroll = scroll;
            format = new DecimalFormat("##.##");
            setAutoscrolls(true);
        }

        public boolean isOpaque() { return true; }
        public boolean isFocusTraversable() { return true; }

        public void paintComponent(Graphics g) {
            int start = getStartTime();
            int finish = getFinishTime();
            Rectangle bounds = getBounds();
            Color c;
            
            // Blank
            g.setColor(Color.white);
            g.fillRect(0, 0, bounds.width, bounds.height);

            // Determine the yscale
            double yscale = (gui.getMaximumYValue() - gui.getMinimumYValue()) / bounds.height;

            // Draw the axis
            g.setColor(Color.black);
            g.drawLine(UNITS_LINE, 0, UNITS_LINE, bounds.height);
            for (int i = bounds.height; i >= 0; i -= UNITS_DIV) {
                g.drawLine(UNITS_LINE, (int)Math.round(i),
                           UNITS_LINE + 2, (int)Math.round(i));
                int valign = (i == 0) ? 2 : (i == bounds.height) ? 0 : 1;
                drawString(g, format.format(((bounds.height - i) * yscale)),
                           UNITS_TEXT, i, 2, valign);
            }
        }

        public Dimension getPreferredSize() {
            return new Dimension(UNITS_WIDTH, 10);
        }
        public Dimension getMinimumSize() { return getPreferredSize(); }
        public Dimension getMaximumSize() { return getPreferredSize(); }
    }

    /**
     * The graph panel
     */
    class GraphCanvas extends JPanel {
        JScrollPane graphscroll;
        GraphGui gui;
        
        public GraphCanvas(GraphGui g, JScrollPane scroll) {
            gui = g;
            graphscroll = scroll;
            setAutoscrolls(true);
        }

        public boolean isOpaque() { return true; }
        public boolean isFocusTraversable() { return true; }

        public void paintComponent(Graphics g) {
            int start = getStartTime();
            int finish = getFinishTime();
            Color c;
            Rectangle bounds = graphscroll.getViewport().getViewRect();
            
            // Limit what we actually draw
            int vstart = calculateTime(bounds.x);
            if (vstart < 0) vstart = start;
            int vfinish = calculateTime(bounds.x + bounds.width);
            if (vfinish < 0) vfinish = finish;

            // Blank
            g.setColor(Color.white);
            g.fillRect(bounds.x, bounds.y, bounds.width, bounds.height);

            // Determine the yscale
            double yscale = bounds.height / (gui.getMaximumYValue() - gui.getMinimumYValue());

            // Draw the data
            g.setColor(Color.red);
            Integer rstart = new Integer(vstart - TIME_GRANULARITY);
            Integer rfinish = new Integer(vfinish + TIME_GRANULARITY);
            Point thispoint = null;
            Point lastpoint = null;
            for (Iterator e = gui.getDataKeys(); e.hasNext(); ) {
                String name = (String)e.next();
                if (!gui.isDataKeyEnabled(name)) continue;
                TreeMap map = (TreeMap)gui.getData().get(name);
                SortedMap range = map.subMap(rstart, rfinish);

                g.setColor(getColor(name));

                lastpoint = null;
                synchronized (map) {
                    for (Iterator i = range.entrySet().iterator(); i.hasNext(); ) {
                        Map.Entry entry = (Map.Entry)i.next();
                        int time = ((Integer)entry.getKey()).intValue();
                        double value = ((Number)entry.getValue()).doubleValue();

                        switch (style) {
                        case GraphGui.GRAPH_COLUMN:
                            thispoint = new Point(Math.round((float)(time - start) * TICKWIDTH),
                                                  bounds.height - (int)Math.round(value * yscale));
                            g.drawLine(thispoint.x, bounds.height, thispoint.x, thispoint.y);
                            break;

                        case GraphGui.GRAPH_LINE:
                            thispoint = new Point(Math.round((float)(time - start) * TICKWIDTH),
                                                  bounds.height - (int)Math.round(value * yscale));
                            if (lastpoint == null) lastpoint = thispoint;
                            g.drawLine(lastpoint.x, lastpoint.y, thispoint.x, thispoint.y);
                            lastpoint = thispoint;
                            break;
                        }
                    }
                }
            }
        }

        public Dimension getPreferredSize() {
            int w = Math.max(Math.round((getFinishTime() - getStartTime()) * TICKWIDTH), 350);
            return new Dimension(w, 100);
        }
        public Dimension getMinimumSize() { return getPreferredSize(); }
        public Dimension getMaximumSize() { return getPreferredSize(); }
    }

    /**
     * This draws the "legend" on the left hand side, indicating
     * the row labels
     */
    class LegendCanvas extends JPanel {
        JScrollPane graphscroll;
        GraphGui gui;

        int LEGEND_SPACE = 15;
        int LEGEND_MARGIN = 3;

        public LegendCanvas(GraphGui g, JScrollPane scroll) {
            gui = g;
            graphscroll = scroll;

            addMouseListener(new MouseAdapter() {
                    public void mouseClicked(MouseEvent e) {
                        String name = findKey(e.getY());
                        if (name != null) {
                            if (gui.isDataKeyEnabled(name)) {
                                gui.disableDataKey(name);
                            } else {
                                gui.enableDataKey(name);
                            }
                        }
                    }
                });
        }

        public boolean isOpaque() { return true; }

        public void paintComponent(Graphics g) {
            int y = LEGEND_SPACE;

            g.setColor(Color.white);
            g.fillRect(0, 0, getWidth(), getHeight());

            g.setColor(Color.black);
            drawString(g, "Legend", LEGEND_MARGIN, y, 0, 0);
            g.drawLine(LEGEND_MARGIN, y, getWidth() - LEGEND_MARGIN, y);
            y += LEGEND_SPACE;
            
            for (Iterator e = gui.getDataKeys(); e.hasNext(); ) {
                String name = (String)e.next();
                g.setColor(getColor(name));
                drawString(g, name, LEGEND_MARGIN, y, 0, 0);
                if (!gui.isDataKeyEnabled(name)) {
                    FontMetrics metrics = g.getFontMetrics();
                    g.drawLine(LEGEND_MARGIN, y - metrics.getHeight() / 2,
                               metrics.stringWidth(name), y - metrics.getHeight() / 2);
                }
                y += LEGEND_SPACE;
            }
        }

        public String findKey(int y) {

            y -= LEGEND_SPACE;
            for (Iterator e = gui.getDataKeys(); y > 0 && e.hasNext(); ) {
                String name = (String)e.next();
                if (y < LEGEND_SPACE)
                    return name;
                y -= LEGEND_SPACE;
            }

            return null;
        }

        public Dimension getPreferredSize() {
            int w = 75;
            FontMetrics metrics = getFontMetrics(getFont());

            for (Iterator e = gui.getDataKeys(); e.hasNext(); ) {
                String name = (String)e.next();
                w = Math.max(w, metrics.stringWidth(name) + LEGEND_MARGIN);
            }

            return new Dimension(w, 100);
        }
        public Dimension getMinimumSize() { return getPreferredSize(); }
        public Dimension getMaximumSize() { return getPreferredSize(); }
    }
}

