package farm.graphcolor.plugins.drivers; 

import java.rmi.*;
import java.rmi.server.*;

import farm.simulator.*;

import java.util.*;
import java.io.*;
import utilities.Log;
import farm.graphcolor.plugins.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
/**
 * The graph color driver, which generates and maintains the graph structure.
 * This includes defining the nodes and edge connectivity, and potentially
 * modifying those values over time.
 * <p>
 * Properties:<dl>
 * <dt><tt>-nodes [integer]</tt>
 * <dd>The number of graph nodes (default 10)
 * <dt><tt>-edges [integer]</tt>
 * <dd>The number of edges (default 15)
 * <dt><tt>-colors [integer]</tt>
 * <dd>The number of colors to allow (default 3)
 * <dt><tt>-type [awc,random,dbo,apo]</tt>
 * <dd>The type of agent to produce (default random)
 * <dt><tt>-file [file name] </tt>
 * <dd>A file used to create the graph (default dynamically created)
 * <dt><tt>-gfile [file name] </tt>
 * <dd>A file used to hold a seed to create the graph (default dynamically created)
 * <dt><tt>-vfile [file name] </tt>
 * <dd>A file used to hold a seed to create the intial assignments(default dynamically created)
 * <dt><tt>-gseed [integer]</tt>
 * <dd>The seed for the graph building random number generator (default random)
 * <dt><tt>-vseed [integer]</tt>
 * <dd>The seed for the initial value random number generator (default random)
 * <dt><tt>-metagents [integer]</tt>
 * <dd>The integer number of metagents to create, used for intial load 
 *     balancing (default 1)
 * <dt><tt>-randomize [sat|any|cluster]</tt>
 * <dd>The type of randomization used to create graph instances: sat creates
 *     satisfiable ones, any creates totally random, cluster creates
 *     cluster type graphs (default sat)
 * <dt><tt>-clusters [integer]</tt>
 * <dd>The number of clusters to create in the cluster method of 
 *     constructing graphs (default 1)
 * <dt><tt>-density [float]</tt>
 * <dd>The density of intra cluster edges 
 *     (default 1.0 all intra cluster)
 * <dt><tt>-maxcost [integer]</tt>
 * <dd>The maximum cost associated with not satisfying a constraint 
 *     (default 1 all constaints equal value)
 * </dl>
 */
public class GraphColorDriver extends FarmPlugIn {

    // Constants
    /**
     * Property identifier for graph nodes
     */
    public static final String CLA_NODES = "nodes";

    /**
     * Property identifier for graph edges
     */
    public static final String CLA_EDGES = "edges";

    /**
     * Property identifier for number of colors in the environemt
     */
    public static final String CLA_COLORS = "colors";

    /**
     * Property identifier for the type of agent to create
     */
    public static final String CLA_TYPE = "type";

    /**
     * Property identifier for the graph seed of the random number generator
     */
    public static final String CLA_GSEED = "gseed";

    /**
     * Property identifier for the initialization seed of the random number generator
     */
    public static final String CLA_VSEED = "vseed";

    /**
     * Property identifier for specifying a file to create the graph
     */
    public static final String CLA_FILE = "file";

    /**
     * Property identifier for specifying a file with a seed to create the graph
     */
    public static final String CLA_GFILE = "gfile";

    /**
     * Property identifier for specifying a file with a seed to create the initial values
     */
    public static final String CLA_VFILE = "vfile";

    /**
     * Property identifier for number of metagents in the environmemt
     */
    public static final String CLA_METAGENTS = "metagents";
    /**
     * Property identifier for number for type of graph to construct
     */
    public static final String CLA_RANDOMIZE = "randomize";

    /**
     * Property identifier for number of clusters in a clusted graph
     */
    public static final String CLA_CLUSTERS = "clusters";

    /**
     * Property identifier for intra cluster edge density
     */
    public static final String CLA_DENSITY = "density";

    /**
     * Property identifier for intra cluster edge density
     */
    public static final String CLA_MAXCOST = "maxcost";

    /**
     * Property identifier for specfying whether the random graph
     * is able to be satisfied or not
     */
    public static final String PROP_SATISFIABLE = "satisfiable";

    /**
     * Property identifier to indicate driver if finshed with initial startup
     */
    public static final String PROP_DRIVERSTARTUP = "dstart";
    
    /**
     * Property identifier for specfying the maximal number of edges
     * that can be satisfied in the random graph
     */
    public static final String PROP_OPTIMAL = "Optimal";

    /**
     * Property identifier for a list of the nodes within each cluster
     */
    public static final String PROP_CLUSTERS = "ClustList";


    int numnodes = 10;
    int numedges = 15;
    int numcolors = 3;
    ArrayList nodes = new ArrayList();
    int metaAgents = 1;
    String type = "random";
    int gseed = 0;
    int vseed = 0;
    String file = "";
    String vfile = "";
    String gfile = "";
    String randomize = "sat";
    boolean satisfiable = true;
    boolean saveGFile = false;
    int numclusters = 1;
    float density = 1.0f;
    int optimal = 0;
    int maxCost = 1;
    
    /**
     * Constructor
     * @param args An array of command line arguments for the driver
     */
    public GraphColorDriver(String []args) throws RemoteException {
	super(args);

        // Command line args
        setPlugInName(getCommandLineArgument(args, CLA_NAME, "Graph Color Driver"));
        setNodes(Integer.valueOf(getCommandLineArgument(args, CLA_NODES, String.valueOf(numnodes))));
        setEdges(Integer.valueOf(getCommandLineArgument(args, CLA_EDGES, String.valueOf(numedges))));
        setColors(Integer.valueOf(getCommandLineArgument(args, CLA_COLORS, String.valueOf(numcolors))));
        setType(getCommandLineArgument(args, CLA_TYPE, type));
        setGSeed(Integer.valueOf(getCommandLineArgument(args, CLA_GSEED, String.valueOf(gseed))));
        setVSeed(Integer.valueOf(getCommandLineArgument(args, CLA_VSEED, String.valueOf(vseed))));
	setFile(getCommandLineArgument(args, CLA_FILE, file));
	setGFile(getCommandLineArgument(args, CLA_GFILE, gfile));
	setVFile(getCommandLineArgument(args, CLA_VFILE, vfile));
	setRandomize(getCommandLineArgument(args, CLA_RANDOMIZE, randomize));
        setClusters(Integer.valueOf(getCommandLineArgument(args, CLA_CLUSTERS, String.valueOf(numclusters))));
        setDensity(Float.valueOf(getCommandLineArgument(args, CLA_DENSITY, String.valueOf(density))));
        setMetaAgents(Integer.valueOf(getCommandLineArgument(args, CLA_METAGENTS, String.valueOf(metaAgents))));
        setMaxCost(Integer.valueOf(getCommandLineArgument(args, CLA_MAXCOST, String.valueOf(maxCost))));

	Random rnd = new Random();
	if(!gfile.equals("")){
	    File f = new File(gfile);
	    if(f.exists()){
		System.out.println("Reading graph seed from file");
		try{
		    BufferedReader infile = new BufferedReader(new FileReader(gfile));
		    setGSeed(new Integer(infile.readLine()));
		    satisfiable = Boolean.valueOf(infile.readLine()).booleanValue();
		    infile.close();
		}catch(IOException e){
		    //ignore it
		}			
		System.out.println("GSeed: " + gseed);
	    }
	
	    else{
		setGSeed(new Integer((int) rnd.nextLong()));
		System.out.println("Creating new graph seed");
		saveGFile = true;
	    }
	    System.out.println("GSeed: " + gseed);
	}
	if(gseed == 0) setGSeed(new Integer((int) rnd.nextLong()));

	if(!vfile.equals("")){
	    File f = new File(vfile);
	    if(f.exists()){
		System.out.println("Reading initial values seed from file");
		try{
		    BufferedReader infile = new BufferedReader(new FileReader(vfile));
		    setVSeed(new Integer(infile.readLine()));
		    infile.close();
		}catch(IOException e){
		    //ignore it
		}				
		System.out.println("VSeed: " + vseed);
	    }
	    else{
		System.out.println("Creating new initial values seed and saving to file");
		setVSeed(new Integer((int) rnd.nextLong()));
		try{
		    PrintWriter outfile= new PrintWriter(new FileWriter(f));
		    outfile.println(vseed);
		    outfile.close();
		}catch(IOException e){
		    //ignore it
		}		
		System.out.println("VSeed: " + vseed);
	    }
	}	
	if(vseed == 0) setVSeed(new Integer((int) rnd.nextLong()));


    }

    /**
     * Sets the number of nodes
     * @param i Number of nodes
     */
    public void setNodes(Integer i) {
        numnodes = i.intValue();
        setProperty(CLA_NODES, i);
    }

    /**
     * Sets the number of edges
     * @param i Number of edges
     */
    public void setEdges(Integer i) {
        numedges = i.intValue();
        setProperty(CLA_EDGES, i);
    }

    /**
     * Sets the number of colors
     * @param i Number of colors
     */
    public void setColors(Integer i) {
        numcolors = i.intValue();
        setProperty(CLA_COLORS, i);
    }

    /**
     * Sets the type of agent to produce
     * @param t The agent type (e.g. random, awc)
     */
    public void setType(String t) {
        type = t;
        setProperty(CLA_TYPE, t);
    }

    /**
     * Sets the name of the file to create the graph from
     * @param f The name of the file
     */
    public void setFile(String f) {
        file = f;
        setProperty(CLA_FILE, f);
    }

    /**
     * Sets the name of the file that holds the seed to create the graph from
     * @param f The name of the file
     */
    public void setGFile(String f) {
        gfile = f;
        setProperty(CLA_GFILE, f);
    }

    /**
     * Sets the name of the file that holds the seed to create an initial assignment from
     * @param f The name of the file
     */
    public void setVFile(String f) {
        vfile = f;
        setProperty(CLA_VFILE, f);
    }

    /**
     * Sets the seed to create the initial values from
     * @param f The name of the file
     */
    public void setVSeed(Integer i) {
        vseed = i.intValue();
        setProperty(CLA_VSEED, i);
    }

    /**
     * Sets the name of the file to create the graph from
     * @param f The name of the file
     */
    public void setGSeed(Integer i) {
        gseed = i.intValue();
	setSeed(i);
        setProperty(CLA_GSEED, i);
    }
    /**
     * Sets the number of metagents
     * @param i The number of metagents
     */
    public void setMetaAgents(Integer i) {
        metaAgents = i.intValue(); 
        setProperty(CLA_METAGENTS, i);
    } 
   
    /**
     * Sets the type of randomized graph that is produced
     * @param f either sat, any, or cluster
     */
    public void setRandomize(String f) {
        randomize = f;
        setProperty(CLA_RANDOMIZE, f);
    }

    /**
     * Sets the number of clusters
     * @param f either sat, any, or cluster
     */
    public void setClusters(Integer f) {
        numclusters = f.intValue();
        setProperty(CLA_CLUSTERS, f);
    }

    /**
     * Sets the intra-cluster and inter-cluster densities
     * @param f 
     */
    public void setDensity(Float f) {
        density = f.floatValue();
        setProperty(CLA_DENSITY, f);
    }

    /**
     * Sets the 
     * @param f 
     */
    public void setMaxCost(Integer i) {
        maxCost = i.intValue();
        setProperty(CLA_MAXCOST, i);
    }

    /**
     * Starts the plugin, creates the sensors, creates the targets,
     * assigns the sensors and targets to metagents
     */
    public void begin() throws RemoteException {
	super.begin();
	if(file.equals(""))
	   nodes = createNodes(numnodes, numedges, randomize);
	else
	    nodes = createNodes(file);
	numnodes = nodes.size();

        for (Iterator i = nodes.iterator(); i.hasNext();) {
            NodeModel n = (NodeModel)i.next();
            log.log("Node: " + n, Log.LOG_INFO);
        }

        // Assign the variables
        log.log("Creating assignments for " + metaAgents + " metagents", Log.LOG_INFO);
        int div = (int)Math.floor(numnodes / metaAgents);
        int mod = numnodes % metaAgents;
        for (int i = metaAgents-1, start = 0; i >= 0; i--) {
            String metkey = "MA" + i  + GraphColorMetAgent.AGENTS_DESCRIPTION;
            int finish = start + div-1;
            if (mod > 0) {
                finish++;
                mod--;
            }
            if (finish >= start) {
                String assign = type + " " + start + "-" + finish;
                String prev = (String)getProperty(metkey);
                if (prev != null)
                    assign = prev + " " + assign;
                setProperty(metkey, assign);
            }
            start = finish + 1;
        }

	setProperty(PROP_DRIVERSTARTUP, new Boolean(true));

    }

    /**
     * Creates the node objects and the edges that connect them
     * from a file specification
     */
    private ArrayList createNodes(String filename){
	ArrayList nodes = new ArrayList();
	try{
	    BufferedReader infile = new BufferedReader(new FileReader(filename));
	    String line;
	    while((line = infile.readLine()) != null){
		System.err.println(line);
		nodes.add(new NodeModel(this, line));
	    }
	    infile.close();
	} catch(IOException e){
	    //ignore it
	}
	setProperty(PROP_SATISFIABLE, new Boolean(true));
	setProperty(PROP_OPTIMAL, new Integer(0));

	Random random = new Random();

	for(Iterator node = nodes.iterator();node.hasNext();){
	    NodeModel model = (NodeModel) node.next();
	    model.setColor(random.nextInt(model.getPossibleColors()));
	    /*if(model.getNumber() == 0 || model.getNumber() == 2 ||
	       model.getNumber() == 1 || model.getNumber() == 3){
		model.setColor(0);
 	    }
	    else{
		model.setColor(1);
		}*/
	}

	long startTime = System.currentTimeMillis();
	optimal = BandB(nodes);
	long endTime = System.currentTimeMillis()- startTime;
	setProperty("OTime", new Integer((int) endTime));
	System.out.println(endTime);
	
	if(optimal == 0) 
	    satisfiable = true;
	else
	    satisfiable = false;
	setProperty(PROP_SATISFIABLE, new Boolean(satisfiable));
	setProperty(PROP_OPTIMAL, new Integer(optimal));


	return nodes;
    
    }


    /**
     * Creates the node objects and the edges to connect them
     */
    private ArrayList createNodes(int numnodes, int numedges, String randomize){
	ArrayList nodes = new ArrayList(numnodes);
        ArrayList edges = new ArrayList(numedges);

	Random rnd = new Random(vseed);

        // Make the nodes
        for (int i = 0; i < numnodes; i++) {
	    NodeModel node = new NodeModel(this, i, rnd.nextInt(numcolors));
	    nodes.add(i, node);
	}
	
	if(randomize.equalsIgnoreCase("sat")){
	    //create a random satisfiable instance
	    satisfiable = true;
	    optimal = 0;
	    createSatGraph(nodes);
	}
	else if(randomize.equalsIgnoreCase("any")){
	    //create a truly randomized instance
	    createRandomGraph(nodes);
	    long startTime = System.currentTimeMillis();
	    optimal = BandB(nodes);
	    long endTime = System.currentTimeMillis()- startTime;
	    setProperty("OTime", new Integer((int) endTime));
	    System.out.println(endTime);

	    if(optimal == 0) 
		satisfiable = true;
	    else
		satisfiable = false;
	}
	else if (randomize.equalsIgnoreCase("cluster")){
	    createClusterGraph(nodes);
	    //optimal = BandB(nodes);
	    if(optimal == 0) 
		satisfiable = true;
	    else
		satisfiable = false;
	}


	setProperty(PROP_SATISFIABLE, new Boolean(satisfiable));
	setProperty(PROP_OPTIMAL, new Integer(optimal));
	if(saveGFile) saveGraphFile();
	return nodes;
    }

    /**
     * Creates a completely random graph
     * @param nodes - an Arraylist of NodeModels to store the results in
     */
    private void createRandomGraph(ArrayList nodes){
	ArrayList edges = new ArrayList(numedges);

        Random random = getRandom();
        log.log("Creating a random graph with " + numnodes + " nodes and " + numedges + " edges", Log.LOG_INFO);
	edges.clear();
	for (int i = 0; i < numedges; i++) {
	    Edge e = null;
	    do {
		int from = random.nextInt(nodes.size());
		int to = random.nextInt(nodes.size());
		if(from != to)
		    e = new Edge(from, to);
	    } while (e == null || edges.contains(e));
	    edges.add(e);
	}

        // Add the edges
        for (Iterator i = edges.iterator(); i.hasNext();) {
	    int cost = random.nextInt(maxCost) + 1;
            Edge e = (Edge)i.next();
            NodeModel from = (NodeModel)nodes.get(e.getFrom());
            NodeModel to = (NodeModel)nodes.get(e.getTo());
            from.addNeighbor(to.getName(), cost);
            to.addNeighbor(from.getName(), cost);
        }
    }

    /**
     * Creates a satisfiable, connected, random graph
     * @param nodes - an Arraylist of NodeModels to store the results in
     */
    private void createSatGraph(ArrayList nodes){
	ArrayList edges = new ArrayList(numedges);
 
        // Make the groups for edge assignment
        ArrayList groups = new ArrayList(numcolors);
        for (int i = 0; i < numcolors; i++) {
            groups.add(i, new ArrayList());
        }
        for (int i = 0; i < numnodes; i++) { // Just store the node number in the group list
            ((ArrayList)groups.get(i % numcolors)).add(new Integer(i));
        }

        // Check
        int maxedges = 0;
        for (int i = 0; i < numcolors; i++) {
            ArrayList l = (ArrayList)groups.get(i);
            maxedges += l.size() * (numnodes - l.size());
        }

        maxedges /= 2;
        if (numedges > maxedges) {
            numedges = maxedges;
            log.log("Warning, reducing excessive number of edges to " + numedges);
        }
        if (numedges < numnodes - 1) {
            numedges = numnodes - 1;
            log.log("Warning, increasing insufficient number of edges to " + numedges);
        }

        // Make the edges
        Random random = getRandom();
        int count = 0;
        log.log("Creating a satisfiable graph with " + numnodes + " nodes and " + numedges + " edges", Log.LOG_INFO);
        do {
            count++;
            edges.clear();
            for (int i = 0; i < numedges; i++) {
                Edge e = null;
                do {
                    int fromgroupindex = random.nextInt(numcolors);
                    int togroupindex = (fromgroupindex + random.nextInt(numcolors - 1) + 1) % numcolors;
                    ArrayList fromgroup = (ArrayList)groups.get(fromgroupindex);
                    ArrayList togroup = (ArrayList)groups.get(togroupindex);

                    if (fromgroup.size() <= 0 || togroup.size() <= 0) continue;
                    int from = ((Integer)fromgroup.get(random.nextInt(fromgroup.size()))).intValue();
                    int to = ((Integer)togroup.get(random.nextInt(togroup.size()))).intValue();
                    e = new Edge(from, to);
                } while (e == null || edges.contains(e));
                edges.add(e);
            }
        } while (!isGraphConnected(nodes, edges));
        log.log("Connected graph found after " + count + " attempts", Log.LOG_INFO);

         // Add the edges
        for (Iterator i = edges.iterator(); i.hasNext();) {
	    int cost = random.nextInt(maxCost) + 1;
            Edge e = (Edge)i.next();
            NodeModel from = (NodeModel)nodes.get(e.getFrom());
            NodeModel to = (NodeModel)nodes.get(e.getTo());
            from.addNeighbor(to.getName(), cost);
            to.addNeighbor(from.getName(), cost);
        }
    }

    /**
     * Creates a graph with n nodes, c clusters, and a intra cluster
     * density of ind, and an intercluster density of outd
     * @param nodes - an Arraylist of NodeModels to store the results in
     */
    private void createClusterGraph(ArrayList nodes){

        // Make the edges
        Random random = getRandom();
        log.log("Creating a clustered graph with " + numnodes + " nodes, " + numclusters + " clusters, " + density + " density", Log.LOG_INFO);
	ArrayList edges = new ArrayList(numedges);

        ArrayList clusters = new ArrayList(numclusters);
        for (int i = 0; i < numclusters; i++) {
            clusters.add(i, new ArrayList());
        }

        for (int i = 0; i < numnodes; i++) {
            ((ArrayList)clusters.get(i % numclusters)).add(new Integer(i));
        }

	setProperty(PROP_CLUSTERS,clusters);

       //create the intra cluster edges
        for (int i = 0; i < Math.floor(numedges*density); i++) {
            Edge e = null;
            do {
                int clusterindex = random.nextInt(numclusters);

                ArrayList cluster = (ArrayList)clusters.get(clusterindex);

                if (cluster.size() <= 0) continue;
                int from = ((Integer)cluster.get(random.nextInt(cluster.size()))).intValue();
                int to = ((Integer)cluster.get(random.nextInt(cluster.size()))).intValue();
                if(to == from)
                    e = null;
                else
                    e = new Edge(from, to);
            } while (e == null || edges.contains(e));
            edges.add(e);
        }

       for (int i = 0; i < Math.ceil(numedges*(1-density)); i++) {
            Edge e = null;
            do {
                int fromclusterindex = random.nextInt(numclusters);
                int toclusterindex = (fromclusterindex + random.nextInt(numclusters - 1) + 1) % numclusters;
                ArrayList fromcluster = (ArrayList)clusters.get(fromclusterindex);
                ArrayList tocluster = (ArrayList)clusters.get(toclusterindex);

                if (fromcluster.size() <= 0 || tocluster.size() <= 0) continue;
                int from = ((Integer)fromcluster.get(random.nextInt(fromcluster.size()))).intValue();
                int to = ((Integer)tocluster.get(random.nextInt(tocluster.size()))).intValue();
                e = new Edge(from, to);
            } while (e == null || edges.contains(e));
            edges.add(e);
        }

	
        // Add the edges
        for (Iterator i = edges.iterator(); i.hasNext();) {
	    int cost = random.nextInt(maxCost) + 1;
            Edge e = (Edge)i.next();
            NodeModel from = (NodeModel)nodes.get(e.getFrom());
            NodeModel to = (NodeModel)nodes.get(e.getTo());
            from.addNeighbor(to.getName(), cost);
            to.addNeighbor(from.getName(), cost);
        }
    }

    private int BandB(ArrayList nodes){
	ArrayList assign = new ArrayList(nodes.size());
	Int bound = new Int(Integer.MAX_VALUE);
	rBandB(nodes, assign, 0, 0, bound);
	System.out.println("Final Bound: " + bound.getValue());
	return bound.getValue();
    }



    /**
     * Does a branch and bound search for the best solution
     * uses backjumping
     * 
     
    private int rBandB(ArrayList nodes, ArrayList assign, 
		       int variable, int count, Int bound, int deepest){


	boolean foundOne = false;
	//is this the bottom of the recursion?
	if(variable == nodes.size()){
	    //all variables assigned, found a better assignment
	    bound.setValue(count);
	    System.out.println("New Bound: " + count);	    
	    return nodes.size() - 1;
	}
	else{
	    //not the bottom, get the next variable
	    NodeModel node = (NodeModel) nodes.get(variable);
	    int colors = node.getPossibleColors();
	    Random rnd = new Random();
	    ArrayList domain = new ArrayList();
	    for(int elem = 0;elem <colors;elem++)
		domain.add(new Integer(elem));
	    
	    int thisDeepest = deepest;
	    for(int elem = 0; elem < colors; elem++){
		//does color satisfy the constraints?
		int color = ((Integer)domain.remove(rnd.nextInt(domain.size()))).intValue();

		//int thisDeepest = deepest;
		int thisCount = 0;
		//System.out.println("Checking color: " + color);
		for(Iterator neigh = node.getNeighbors();neigh.hasNext();){
		    String neighName = (String) neigh.next();
		    int neighNode = Integer.parseInt((neighName).substring(2));
		    if(neighNode < variable){
			int val = ((Integer) assign.get(neighNode))
			    .intValue();
			if(val == color){
			    thisCount += node.getNeighborValue(neighName);
			    if(neighNode > thisDeepest){
			    //System.out.println(variable + " is deeper than " + thisDeepest);
				thisDeepest = neighNode;
			    }
			}
		    }
		} 

		//assign.add(variable, new Integer(color));
		//System.out.println("Node: " + node.getName() + " count " + (thisCount + count) + " bound " + bound.getValue() + " partial " + assign.subList(0,variable+1).toString());

		if(count + thisCount < bound.getValue()){
		    assign.add(variable, new Integer(color));
		    foundOne = true;
		    int viol = rBandB(nodes, assign, variable + 1, count + thisCount, bound, thisDeepest);
		    if(viol != node.getNumber()){
			//System.out.println("Node " + variable +": not me, jumping to " + viol);
			return viol;
		    }
		}
	    }
	    //all colors checked
	    if(foundOne){
	       //System.out.println("Node " + variable + ": Checked all values, going to "+ (variable -1));
	       return variable -1;
	    }
	    else{
		//System.out.println("Node " + variable + ": Checked all values, going to "+ deepest);
		return thisDeepest;
	    }
	}
    }
    */

    /**
     * A barnch and bound search using back tracking
     */
    private void rBandB(ArrayList nodes, ArrayList assign, 
		       int variable, int count, Int bound){


	//is this the bottom of the recursion?
	if(variable == nodes.size()){
	    //all variables assigned, found a better assignment
	    bound.setValue(count);
	    System.out.println("New Bound: " + count);	    
	    return;
	}
	else{
	    //not the bottom, get the next variable
	    NodeModel node = (NodeModel) nodes.get(variable);
	    //get the domain and randomize it
	    int colors = node.getPossibleColors();
	    Random rnd = new Random();
	    ArrayList domain = new ArrayList();
	    for(int elem = 0;elem <colors;elem++)
		domain.add(new Integer(elem));
	    
	    for(int elem = 0; elem < colors; elem++){
		//does color satisfy the constraints?
		int color = ((Integer)domain.remove(rnd.nextInt(domain.size()))).intValue();

		int thisCount = 0;
		//System.out.println("Checking color: " + color);
		for(Iterator neigh = node.getNeighbors();neigh.hasNext();){
		    String neighName = (String) neigh.next();
		    int neighNode = Integer.parseInt((neighName).substring(2));
		    if(neighNode < variable){
			int val = ((Integer) assign.get(neighNode))
			    .intValue();
			if(val == color)
			    thisCount += node.getNeighborValue(neighName);
		    }
		} 
		
		//assign.add(variable, new Integer(color));
		//System.out.println("Node: " + node.getName() + " count " + (thisCount + count) + " bound " + bound.getValue() + " partial " + assign.subList(0,variable+1).toString());
		
		if(count + thisCount < bound.getValue()){
		    assign.add(variable, new Integer(color));
		    rBandB(nodes, assign, variable + 1, count + thisCount, bound);
		}
		
	    }
	}
    }
    


    /**
     * Given a n-coloring instance, determine if the graph is satisfiable
     * @param nodes - and arry list of node models decribing the problem
     * @return true if satisfiable, false otherwise
     */
    private boolean isSatisfiable(ArrayList nodes){
        log.log("Checking graph satisfiablity", Log.LOG_INFO);
	ArrayList assign = new ArrayList(nodes.size());
	HashSet conflictSet = new HashSet(nodes.size());
	if(rSat(nodes, 0, assign, conflictSet)){
	    log.log("Graph is satisfiable", Log.LOG_INFO);
	    return true;
	}
	else{
	    log.log("Graph is unsatisfiable", Log.LOG_INFO);
	    return false;
	}
    }


    /**
     * Recursive backjumping search for n-graph coloring
     * @param nodes - and ArrayList of NodeModels decribing the graph
     * @param variable - the current variable number
     * @param assign - the currently assigned values in the search
     * @param conflictSet - the current conflictSet
     * @return true if an assignment is found, false otherwise
     */
    private boolean rSat(ArrayList nodes, int variable, ArrayList assign, HashSet conflictSet){
    
	//is this the bottom of the recursion?
	if(variable == nodes.size()){
	    //all variables assigned, found a true assignment
	    return true;
	}
	else{
	    //not the bottom, get the next variable
	    NodeModel node = (NodeModel) nodes.get(variable);
	    //System.out.println("Node: " + node.toString());

	    int colors = node.getPossibleColors();
	    boolean answer = false;
	    conflictSet.clear();
	    for(int color = 0; color < colors; color++){
		//does color satisfy the constraints?
		String good = "";
		//System.out.println("Checking color: " + color);
		for(Iterator neigh = node.getNeighbors();neigh.hasNext();){
		    String neighName = (String) neigh.next();
		    int neighNode = Integer.parseInt((neighName).substring(2));
		    if(neighNode < variable){
			int val = ((Integer) assign.get(neighNode))
			    .intValue();
			if(val == color){
			    good = neighName;
			    break;
			}
		    }
		}
		HashSet newConflicts = new HashSet();
		if(good.equals("")){
		    //satisfies all constraints
		    //System.out.println("Color has no conflicts");
		    assign.add(variable, new Integer(color));
		    
		    answer = rSat(nodes, variable+1, assign, newConflicts);
		}
		else{
		    newConflicts.add(good);
		    newConflicts.add(node.getName());
		}

		if(answer){
		    conflictSet.clear();
		    return answer;
		}
		else if(!newConflicts.contains(node.getName())){
		    conflictSet.addAll(newConflicts);
		    //System.out.println("Node: " + node.getName() + " popping because I'm not in " + conflictSet.toString());
		    return false;
		}
		else{
		    newConflicts.remove(node.getName());
		    //System.out.println("Node: " + node.getName() + " adding new conflicts to the set " + newConflicts.toString());		    
		    conflictSet.addAll(newConflicts);
		}
	    }
	    return false;
	}
    }

    /**
     * Determines if a graph is connected
     * @param nodes ArrayList of NodeModels
     * @param edges ArrayList of strings "ND0 ND5"
     */
    private boolean isGraphConnected(ArrayList nodes, ArrayList edges) {
        ArrayList open = (ArrayList)nodes.clone();
        ArrayList search = new ArrayList();

        log.log("Checking graph connectivity", Log.LOG_INFO);

        // Make the adjacency matrix
        boolean adj[][] = new boolean[nodes.size()][nodes.size()];
        for (int i = 0; i < adj.length; i++) {
            for (int j = 0; j < adj[0].length; j++) {
                adj[i][j] = (i == j);
            }
        }
        // Add the edges
        for (Iterator i = edges.iterator(); i.hasNext();) {
            Edge e = (Edge)i.next();
            adj[e.getFrom()][e.getTo()] = true;
            adj[e.getTo()][e.getFrom()] = true;
        }
        /*for (int i = 0; i < adj.length; i++) {
            for (int j = 0; j < adj[0].length; j++) {
                System.err.print(adj[i][j] ? "1" : "0");
            }
            System.err.println("");
            }*/

        // Get the starting node
        NodeModel n = (NodeModel)nodes.get(0);
        search.add(n);
        open.remove(n);

        // Keep on searching
        while (!search.isEmpty()) {
            n = (NodeModel)search.remove(0);
            for (int i = 0; i < adj.length; i++) {
                if (adj[n.getNumber()][i] == true &&
                    open.contains(nodes.get(i))) {
                    NodeModel m = (NodeModel)nodes.get(i);
                    search.add(open.remove(open.indexOf(m)));
                }
            }
        }
        
        return open.isEmpty();
    }

    private void saveGraphFile(){
	File f = new File(gfile);
	try{
	    PrintWriter outfile = new PrintWriter(new FileWriter(f));
	    outfile.println(gseed);
	    outfile.println(satisfiable);
	    outfile.close();
	}catch(IOException e){
	    //ignore it
	}		
   }

    /**
     * Main Function
     */
    public static void main(String[] args) {
	try {
	    FarmPlugIn plugin = new GraphColorDriver(args);
	    plugin.run();
	} catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }    

}

class Edge {
    int from, to;

    public Edge(int f, int t) {
        from = f;
        to = t;
    }
    public int getFrom() { return from; }
    public int getTo() { return to; }
    public boolean equals(Object o) {
        Edge e = (Edge)o;
        if (getFrom() == e.getFrom() && getTo() == e.getTo()) return true;
        if (getTo() == e.getFrom() && getFrom() == e.getTo()) return true;
        return false;
    }
    public String toString() {
        return "" + getFrom() + "-" + getTo();
    }
}

class Int{

    int value;
    Int(int val){
	setValue(val);
    }
    
    Int(){
	setValue(0);
    }
    
    public void setValue(int val){
	value = val;
    }
    
    public int getValue(){return value;}
}


class Str{

    String value;
    Str(String val){
	setValue(val);
    }
    
    Str(){
	setValue("");
    }
    
    public void setValue(String val){
	value = val;
    }
    
    public String getValue(){return value;}

}
